from concurrent.futures import ThreadPoolExecutor, wait
from typing import Any, Callable, Optional

from tqdm import tqdm


def multithread_run(
    func: Callable,
    input_kwargs: list[dict[str, Any]],
    n_threads: Optional[int] = None,
) -> list[Any]:
    """Run the function with the arguments.

    Args:
        func (Callable): The function to be executed.
        input_kwargs (list[dict[str, Any]]): The input arguments.
        n_threads (Optional[int]): The number of threads to be used.

    Returns:
        results (list[Any]): The results of the function.
    """

    with ThreadPoolExecutor(max_workers=n_threads) as exc:
        futures = [exc.submit(func, **kwargs) for kwargs in input_kwargs]
        wait(tqdm(futures))

    results = [future.result() for future in futures]

    return results



import pandas as pd

from archeo.utils.file import check_and_create_dir
from archeo.visualization.distribution import distribution_summary, kick_against_spin_cmap, kick_distribution_on_spin
from archeo.visualization.estimation import (
    corner_estimates,
    effective_spin_estimates,
    mass_estimates,
    precession_spin_estimates,
    second_generation_probability_curve,
    table_estimates,
)


def visualize_prior_distribution(prior: pd.DataFrame, output_dir: str, fmt: str = "png") -> None:
    """Visualize the prior distribution of the mass, spin, and precession of the black holes.

    Args:
        prior (pd.DataFrame): The prior samples.
        output_dir (str): The directory where the visualizations will be saved.
        fmt (str): The format of the visualizations. Defaults to "png".
    """

    check_and_create_dir(output_dir)

    distribution_summary(prior, output_dir=output_dir, fmt=fmt)
    kick_against_spin_cmap(prior, output_dir=output_dir, fmt=fmt)
    kick_distribution_on_spin(prior, output_dir=output_dir, fmt=fmt)


def visualize_posterior_estimation(
    dfs: dict[str, pd.DataFrame],
    output_dir: str,
    fmt: str = "png",
) -> None:
    """Visualize the posterior estimation of the mass, spin, and precession of the black holes.

    Args:
        dfs (dict[str, pd.DataFrame]): A dictionary of dataframes with the posterior samples.
        output_dir (str): The directory where the visualizations will be saved.
        fmt (str): The format of the visualizations. Defaults to "png".
    """

    check_and_create_dir(output_dir)

    for label, df in dfs.items():
        _output_dir = f"{output_dir}/{label}"
        check_and_create_dir(_output_dir)
        mass_estimates(df, label, output_dir=_output_dir, fmt=fmt)
        corner_estimates({label: df}, output_dir=_output_dir, fmt=fmt)
        second_generation_probability_curve({label: df}, output_dir=_output_dir, fmt=fmt)
        effective_spin_estimates({label: df}, output_dir=_output_dir, fmt=fmt)
        precession_spin_estimates({label: df}, output_dir=_output_dir, fmt=fmt)

    corner_estimates(dfs, output_dir=output_dir, fmt=fmt)
    second_generation_probability_curve(dfs, output_dir=output_dir, fmt=fmt)
    effective_spin_estimates(dfs, output_dir=output_dir, fmt=fmt)
    precession_spin_estimates(dfs, output_dir=output_dir, fmt=fmt)
    table_estimates(dfs, output_dir=output_dir, fmt="md")



import numpy as np

import archeo.logger


local_logger = archeo.logger.get_logger(__name__)


def sph2cart(theta: float, phi: float) -> np.ndarray:
    """Convert spherical coordinates to Cartesian coordinates.

    Args:
        theta (float): Polar angle.
        phi (float): Azimuthal angle.

    Returns:
        cartesian_vector (np.ndarray): Cartesian coordinates.
    """

    cartesian_vector = [
        np.sin(theta) * np.cos(phi),
        np.sin(theta) * np.sin(phi),
        np.cos(theta),
    ]
    return np.array(cartesian_vector, dtype=float)



import json
from dataclasses import asdict, dataclass

import numpy as np
import yaml

from archeo.constants import Fits


@dataclass(frozen=True)
class Domain:
    """Domain of a parameter.

    Attributes:
        low (float): Lower bound of the domain
        high (float): Upper bound of the domain
    """

    low: float = float("-inf")
    high: float = float("inf")

    def contain(self, value: float) -> bool:
        """Check if the input value is within the domain."""

        return self.low <= value <= self.high

    def draw(self) -> float:
        """Draw a random value from the domain."""

        return np.random.uniform(low=self.low, high=self.high)


@dataclass(frozen=True)
class Binary:
    """Binary parameters.

    Attributes:
        m_1 (float): Mass of the primary black hole (in solar mass)
        m_2 (float): Mass of the secondary black hole (in solar mass)
        chi_1 (tuple[float, float, float]): Spin of the primary black hole (dimensionless), [0, 1]
        chi_2 (tuple[float, float, float]): Spin of the secondary black hole (dimensionless), [0, 1]
    """

    m_1: float
    m_2: float
    chi_1: tuple[float, float, float]
    chi_2: tuple[float, float, float]


@dataclass(frozen=True)
class Event:
    """Event (Binary Black Hole Merger) parameters."""

    m_1: float
    m_2: float
    m_ret: float
    m_ret_err: float
    v_f: tuple[float, float, float]
    v_f_err: tuple[float, float, float]
    chi_1: tuple[float, float, float]
    chi_2: tuple[float, float, float]
    chi_f: tuple[float, float, float]
    chi_f_err: tuple[float, float, float]


@dataclass(frozen=True)
class PriorConfig:
    """Configuration of the prior.

    Attributes:
        n_samples (int): Number of samples to generate.
        fits (Fits): Surrogate model to use.
        is_spin_aligned (bool): Whether the spins are aligned or not.
        is_only_up_aligned_spin (bool): Whether the spins are only in the positive z-direction.
        spin (Domain): Domain of the spin parameter.
        phi (Domain): Domain of the azimuthal angle of the spin.
        theta (Domain): Domain of the polar angle of the spin.
        mass_ratio (Domain): Domain of the mass ratio.
        mass (Domain): Domain of the mass.
        is_mahapatra (bool): Whether the Mahapatra mass function is used.
    """

    n_samples: int
    fits: Fits
    is_spin_aligned: bool
    is_only_up_aligned_spin: bool
    a_1: Domain  # unit: dimensionless
    a_2: Domain  # unit: dimensionless
    phi_1: Domain  # unit: pi
    phi_2: Domain  # unit: pi
    theta_1: Domain  # unit: pi
    theta_2: Domain  # unit: pi
    mass_ratio: Domain  # unit: dimensionless
    m_1: Domain  # unit: solar mass
    m_2: Domain  # unit: solar mass
    is_mahapatra: bool = False
    is_uniform_in_mass_ratio: bool = False

    def __post_init__(self) -> None:
        """Post initialization."""

        self._check_fits_is_valid()
        self._check_is_aligned_spin_if_only_up_aligned_spin()
        self._check_is_m1_highbound_greater_than_m2_lowbound()

    def _check_fits_is_valid(self) -> None:
        """Check if the surrogate model is valid."""

        if self.fits not in Fits:
            raise ValueError(f"Invalid fits: {self.fits}")

    def _check_is_aligned_spin_if_only_up_aligned_spin(self) -> None:
        """Check if spins are aligned if only up-aligned spin is set."""

        if self.is_only_up_aligned_spin and not self.is_spin_aligned:
            raise ValueError("Only up-aligned spin is only valid when spins are aligned.")

    def _check_is_m1_highbound_greater_than_m2_lowbound(self) -> None:
        """Check if the high bound of m_1 is greater than the low bound of m_2."""

        if self.m_1.high <= self.m_2.low:
            raise ValueError("The high bound of m_1 must be greater than the low bound of m_2.")

    @staticmethod
    def from_dict(data: dict) -> "PriorConfig":
        """Create a PriorConfig object from a dictionary."""

        return PriorConfig(
            n_samples=data["n_samples"],
            fits=Fits(data["fits"]),
            is_spin_aligned=data["is_spin_aligned"],
            is_only_up_aligned_spin=data["is_only_up_aligned_spin"],
            a_1=Domain(data["a_1"]["low"], data["a_1"]["high"]),
            a_2=Domain(data["a_2"]["low"], data["a_2"]["high"]),
            phi_1=Domain(data["phi_1"]["low"], data["phi_1"]["high"]),
            phi_2=Domain(data["phi_2"]["low"], data["phi_2"]["high"]),
            theta_1=Domain(data["theta_1"]["low"], data["theta_1"]["high"]),
            theta_2=Domain(data["theta_2"]["low"], data["theta_2"]["high"]),
            mass_ratio=Domain(data["mass_ratio"]["low"], data["mass_ratio"]["high"]),
            m_1=Domain(data["m_1"]["low"], data["m_1"]["high"]),
            m_2=Domain(data["m_2"]["low"], data["m_2"]["high"]),
            is_mahapatra=data["is_mahapatra"],
            is_uniform_in_mass_ratio=data["is_uniform_in_mass_ratio"],
        )

    @classmethod
    def from_json(cls, filepath: str) -> "PriorConfig":
        """Create a PriorConfig object from a JSON file."""

        with open(filepath, "r", encoding="utf-8") as f:
            data = json.load(f)

        return cls.from_dict(data)

    @classmethod
    def from_yaml(cls, filepath: str) -> "PriorConfig":
        """Create a PriorConfig object from a YAML file."""

        with open(filepath, "r", encoding="utf-8") as f:
            data = yaml.safe_load(f)

        return cls.from_dict(data)

    def to_dict(self) -> dict:
        """Convert to a dictionary."""

        return asdict(self)

    def to_json(self, filepath: str) -> None:
        """Save to a JSON file."""

        with open(filepath, "w", encoding="utf-8") as f:
            json.dump(self.to_dict(), f, indent=4)

    def to_yaml(self, filepath: str) -> None:
        """Save to a YAML file."""

        with open(filepath, "w", encoding="utf-8") as f:
            yaml.dump(self.to_dict(), f)


@dataclass(frozen=True)
class Padding:
    """
    Padding for plot.

    Attributes:
        tpad: float, the top padding of the plot.
        lpad: float, the left padding of the plot.
        bpad: float, the bottom padding of the plot.
    """

    tpad: float = 2.5
    lpad: float = 0.1
    bpad: float = 0.12


@dataclass(frozen=True)
class Labels:
    """
    Labels for plot.

    Attributes:
        title: str, the title of the plot.
        xlabel: str, the x-axis label of the plot.
        ylabel: str, the y-axis label of the plot.
    """

    title: str = ""
    xlabel: str = ""
    ylabel: str = ""



from typing import Callable

from archeo.logger import get_logger


local_logger = get_logger(__name__)


def pre_release(func: Callable) -> Callable:
    """Decorator to mark a function as a pre-release feature."""

    def wrapper(*args, **kwargs):
        """Wrapper function to log a warning message before executing the decorated function."""

        local_logger.warning(
            "%s is a pre-release feature. Correctness is not guaranteed.",
            func.__name__,
        )
        return func(*args, **kwargs)

    return wrapper



import numpy as np
import pandas as pd
import seaborn as sns
from scipy import interpolate

from archeo.core.forward.bayes import ISData, get_bayes_factor
from archeo.utils.helper import pre_release
from archeo.visualization.base import add_escape_velocity


@pre_release
def plot_bayes_factor_over_kick(
    ax,
    prior: pd.DataFrame,
    posterior: pd.DataFrame,
    candidate_prior: dict[str, pd.Series],
    kicks: pd.Series,
    label: str,
    n_bounds: int = 100,
    min_nk: int = 1000,
    show_escape_velocity: bool = False,
    show_scatter: bool = False,
):
    """Plot the Bayes factor over the kick range.

    Args:
        ax: The axis to plot on.
        prior: The prior DataFrame.
        posterior: The posterior DataFrame.
        candidate_prior: The candidate prior DataFrame.
        kicks: The kick values.
        label: The label of the plot.
        n_bounds: The number of bounds to interpolate.
        min_nk: The minimum number of kicks to consider.
        show_escape_velocity: Whether to show the escape velocity.
        show_scatter: Whether to show the scatter plot.

    Returns:
        The axis with the plot.
    """

    k_lb = kicks.quantile(min_nk / len(kicks))
    k_ub = kicks.max()

    _ks = np.logspace(np.log10(k_lb), np.log10(k_ub), n_bounds)
    bfs: list[float] = []

    for k in _ks:
        # Mark index of the samples that are below the kick threshold
        idx = kicks < k
        _candidate_prior = {col: ds[idx] for col, ds in candidate_prior.items()}
        data = ISData(candidate_prior=_candidate_prior, posterior=posterior, prior=prior)
        bfs.append(get_bayes_factor(data))

    f = interpolate.interp1d(_ks, bfs, fill_value="extrapolate")
    x = np.linspace(k_lb, k_ub, 1000)
    y = f(x)

    sns.lineplot(y=y, x=x, ax=ax, label=label)
    if show_scatter:
        sns.scatterplot(y=bfs, x=_ks, ax=ax, color="black")

    if show_escape_velocity:
        add_escape_velocity(ax, k_lb, max(bfs))

    return ax



import os

import pandas as pd

import archeo.logger


local_logger = archeo.logger.get_logger(__name__)


def check_and_create_dir(dirpath: str) -> None:
    """Check if the directory exists, if not create it.
    NOTE: We recursively create the directory if it does not exist.

    Args:
        dirpath (str): The directory path
    """

    if not os.path.exists(dirpath):
        os.makedirs(dirpath, exist_ok=True)

    local_logger.info("Created directory: %s", dirpath)


def read_data(filepath: str) -> pd.DataFrame:
    """Read data from a file according to the file extension.

    Args:
        filepath (str): The file path

    Returns:
        pd.DataFrame: The data in a pandas DataFrame
    """

    filepath = filepath.lower()

    if filepath.endswith(".feather") or filepath.endswith(".ipc"):
        return pd.read_feather(filepath)

    if filepath.endswith(".csv"):
        return pd.read_csv(filepath)

    if filepath.endswith(".parquet"):
        return pd.read_parquet(filepath)

    if filepath.endswith(".json"):
        return pd.read_json(filepath)

    raise ValueError(f"File extension not supported: {filepath}")



from dataclasses import dataclass

import numpy as np
import pandas as pd
from scipy.stats import rv_histogram

import archeo.logger
from archeo.schema import Domain
from archeo.utils.helper import pre_release


local_logger = archeo.logger.get_logger(__name__)


@dataclass
class ISData:
    """Importance sampling data"""

    posterior: pd.DataFrame
    prior: pd.DataFrame
    candidate_prior: dict[str, pd.Series]

    def _check_column_existence(self) -> None:
        """Check if the columns in the dataframes are consistent"""

        for col in self.candidate_prior.keys():
            if col not in self.prior.columns:
                raise ValueError(f"Column {col} is not in the prior dataframe")
            if col not in self.posterior.columns:
                raise ValueError(f"Column {col} is not in the posterior dataframe")

        for col in self.prior.columns:
            if col not in self.posterior.columns:
                raise ValueError(f"Prior column {col} is not in the posterior dataframe")

        for col in self.posterior.columns:
            if col not in self.prior.columns:
                raise ValueError(f"Posterior column {col} is not in the prior dataframe")

    def _compute_bounds(self) -> dict[str, Domain]:
        """Get the bounds of the dataframes"""

        bounds: dict[str, Domain] = {}

        for col in self.posterior.columns:
            _min = min(
                self.posterior[col].min(),
                self.prior[col].min(),
                (self.candidate_prior[col].min() if col in self.candidate_prior else np.inf),
            )
            _max = max(
                self.posterior[col].max(),
                self.prior[col].max(),
                (self.candidate_prior[col].max() if col in self.candidate_prior else -np.inf),
            )

            bounds[col] = Domain(low=_min, high=_max)

        return bounds

    def __post_init__(self) -> None:
        """Check if the columns in the dataframes are consistent"""

        self._check_column_existence()
        self._bounds = self._compute_bounds()

    @property
    def bounds(self) -> dict[str, Domain]:
        """Get the bounds of the dataframes"""

        return self._bounds

    @property
    def correlated_params(self) -> list[str]:
        """Get the correlated parameters"""

        return [col for col in self.prior.columns if col not in self.candidate_prior]


def get_histrogram(samples: pd.Series, bounds: Domain, nbins: int = 70, atol: float = 1e-6) -> np.ndarray:
    """Compute the histogram of the samples"""

    hist, edges = np.histogram(samples, bins=nbins, density=True, range=(bounds.low, bounds.high))
    binwidth = edges[1] - edges[0]

    auc = np.sum(hist) * binwidth
    if not np.isclose(auc, 1.0, atol=atol):
        msg = f"Invalid probability distribution (AUC={auc:.2f})."
        local_logger.error(msg)
        raise ValueError(msg)

    return hist


@pre_release
def get_weights(data: ISData, nbins=70, ztol=1e-8) -> np.ndarray:
    """Get the weights for the importance sampling

    Args:
        data (ISData): Importance sampling data.
        nbins (int, optional): Number of bins for the histogram. Defaults to 70.
        ztol (float, optional): Tolerance for the zero division. Defaults to 1e-8.

    Returns:
        weights: np.ndarray, the weights for the importance sampling.
    """

    weights = np.ones(len(data.posterior))

    for col, samples in data.candidate_prior.items():
        prior_hist = get_histrogram(data.prior[col], data.bounds[col], nbins)
        candidate_prior_hist = get_histrogram(samples, data.bounds[col], nbins)

        # Avoid division by zero
        ratio = np.where(prior_hist > ztol, np.exp(np.log(candidate_prior_hist) - np.log(prior_hist)), 0.0)
        rv = rv_histogram((ratio, np.linspace(data.bounds[col].low, data.bounds[col].high, nbins + 1)))
        weights *= rv.pdf(data.posterior[col])

    return weights


@pre_release
def get_bayes_factor(data: ISData, nbins=70, random_state=42, ztol: float = 1e-8) -> float:
    """Compute the Bayes factor between two models

    NOTE: In this implementation, the likelihood function remains untouched.
    So that the Bayes factor is computed as the ratio of the new prior to the old prior.
    Details please check importance sampling.

    Args:
        data (ISData): Importance sampling data.
        nbins (int, optional): Number of bins for the histogram. Defaults to 70.
        random_state (int, optional): Random state for resampling the prior. Defaults to 42.
        ztol (float, optional): Tolerance for the zero division. Defaults to 1e-8.

    Returns:
        bf: float, the Bayes factor.
    """

    bf = 1.0
    weights = np.ones(len(data.prior))

    # Here we first loop over the candidate prior columns
    # We need to compute the weights for resampling the prior
    # for the correction of potentially correlated parameters
    for col, samples in data.candidate_prior.items():
        prior_hist = get_histrogram(data.prior[col], data.bounds[col], nbins)
        posterior_hist = get_histrogram(data.posterior[col], data.bounds[col], nbins)
        candidate_prior_hist = get_histrogram(samples, data.bounds[col], nbins)
        binwidth = (data.bounds[col].high - data.bounds[col].low) / nbins

        ratio = np.where(prior_hist > ztol, np.exp(np.log(candidate_prior_hist) - np.log(prior_hist)), 0.0)
        rv = rv_histogram((ratio, np.linspace(data.bounds[col].low, data.bounds[col].high, nbins + 1)))
        weights *= rv.pdf(data.prior[col])

        bf *= (
            np.sum(
                np.where(
                    prior_hist > ztol,
                    np.exp(np.log(candidate_prior_hist) + np.log(posterior_hist) - np.log(prior_hist)),
                    0.0,
                )
            )
            * binwidth
        )

    correlated_prior = data.prior.sample(n=len(data.prior), weights=weights, replace=True, random_state=random_state)
    # Here we loop over the potentially correlated parameters
    # to compute the Bayes factor.
    for col in data.correlated_params:
        prior_hist = get_histrogram(data.prior[col], data.bounds[col], nbins)
        posterior_hist = get_histrogram(data.posterior[col], data.bounds[col], nbins)
        candidate_prior_hist = get_histrogram(correlated_prior[col], data.bounds[col], nbins)
        binwidth = (data.bounds[col].high - data.bounds[col].low) / nbins

        bf *= (
            np.sum(
                np.where(
                    prior_hist > ztol,
                    np.exp(np.log(candidate_prior_hist) + np.log(posterior_hist) - np.log(prior_hist)),
                    0.0,
                )
            )
            * binwidth
        )

    return bf
